﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

using OpenTK.Graphics.OpenGL;
using OpenTK;

namespace OpenTkDrawTank
{
    public partial class Form1 : Form
    {
        Tank Tank = new Tank();

        public Form1()
        {
            InitializeComponent();

            ShowLabel();
            Tank.Color = Color.Pink;
        }

        // 視点の X座標
        float _eyeX = 0;
        float EyeX
        {
            get { return _eyeX; }
            set
            {
                _eyeX = (float)Math.Round(value, 2, MidpointRounding.AwayFromZero);
                ShowLabel();
                SetSight();
            }
        }

        // 視点の Y座標
        float _eyeY = 0;
        float EyeY
        {
            get { return _eyeY; }
            set
            {
                _eyeY = (float)Math.Round(value, 2, MidpointRounding.AwayFromZero);
                ShowLabel();
                SetSight();
            }
        }

        // 視点の Z座標
        float _eyeZ = 0;
        float EyeZ
        {
            get { return _eyeZ; }
            set
            {
                _eyeZ = (float)Math.Round(value, 2, MidpointRounding.AwayFromZero);
                ShowLabel();
                SetSight();
            }
        }

        // 視線のターゲット X座標
        float _targetX = 0;
        float TargetX
        {
            get { return _targetX; }
            set
            {
                _targetX = (float)Math.Round(value, 2, MidpointRounding.AwayFromZero);
                ShowLabel();
                SetSight();
            }
        }

        // 視線のターゲット Y座標
        float _targetY = 0;
        float TargetY
        {
            get { return _targetY; }
            set
            {
                _targetY = (float)Math.Round(value, 2, MidpointRounding.AwayFromZero);
                ShowLabel();
                SetSight();
            }
        }

        // 視線のターゲット Z座標
        float _targetZ = 0;
        float TargetZ
        {
            get { return _targetZ; }
            set
            {
                _targetZ = (float)Math.Round(value, 2, MidpointRounding.AwayFromZero);
                ShowLabel();
                SetSight();
            }
        }

        void SetSight()
        {
            // 視界の設定
            Vector3 eye = new Vector3(EyeX, EyeY, EyeZ);
            Vector3 target = new Vector3(TargetX, TargetY, TargetZ);
            Vector3 up = new Vector3(0, 1, 0);

            Matrix4 look = Matrix4.LookAt(eye, target, up);
            GL.LoadMatrix(ref look);
        }

        void ShowLabel()
        {
            labelRotateX.Text = RotateX.ToString();
            labelRotateY.Text = RotateY.ToString();
            labelRotateZ.Text = RotateZ.ToString();

            labelEyeX.Text = EyeX.ToString();
            labelEyeY.Text = EyeY.ToString();
            labelEyeZ.Text = EyeZ.ToString();

            labelTargetX.Text = TargetX.ToString();
            labelTargetY.Text = TargetY.ToString();
            labelTargetZ.Text = TargetZ.ToString();
        }

        private void glControlEx1_Load(object sender, EventArgs e)
        {
            Init();
        }

        void Init()
        {
            GL.ClearColor(glControl.BackColor);

            // Projection の設定
            SetProjection();

            // デプスバッファの使用
            GL.Enable(EnableCap.DepthTest);

            // 法線の自動調節
            GL.Enable(EnableCap.Normalize);

            EyeX = 0f;
            EyeY = 1f;
            EyeZ = 2f;

            Lighting();
            glControl.Refresh();
        }

        private void SetProjection()
        {
            // ビューポートの設定
            GL.Viewport(0, 0, glControl.Width, glControl.Height);

            // 視体積の設定
            GL.MatrixMode(MatrixMode.Projection);
            Matrix4 proj = Matrix4.CreatePerspectiveFieldOfView(MathHelper.PiOver3, glControl.AspectRatio, 0.01f, 30.0f);
            GL.LoadMatrix(ref proj);

            // MatrixMode を元に戻す
            GL.MatrixMode(MatrixMode.Modelview);
        }

        void Lighting()
        {
            // 光源の使用
            GL.Enable(EnableCap.Lighting);

            // ライト 0 の設定と使用
            float[] position = new float[] { 0.8f, 1.0f, 1.2f, 0.0f };
            GL.Light(LightName.Light0, LightParameter.Position, position);

            GL.Enable(EnableCap.Light0);
        }

        private void glControl_Resize(object sender, EventArgs e)
        {
            SetProjection();

            // 再描画
            glControl.Refresh();
        }

        private void glControl_Paint(object sender, PaintEventArgs e)
        {
            GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);

            // 視点を変更する
            SetSight();

            Tank.Draw();

            glControl.SwapBuffers();
        }

        // X軸を中心とした回転角度
        float RotateX
        {
            get { return Tank.RotateX; }
            set
            {
                if (value < 0)
                    value += 360;
                Tank.RotateX = value % 360;

                ShowLabel();
                SetSight();
            }
        }

        // Y軸を中心とした回転角度
        float RotateY
        {
            get { return Tank.RotateY; }
            set
            {
                if (value < 0)
                    value += 360;
                Tank.RotateY = value % 360;

                ShowLabel();
                SetSight();
            }
        }

        // Z軸を中心とした回転角度
        float RotateZ
        {
            get { return Tank.RotateZ; }
            set
            {
                if (value < 0)
                    value += 360;
                Tank.RotateZ = value % 360;

                ShowLabel();
                SetSight();
            }
        }

        protected override bool ProcessDialogKey(Keys keyData)
        {
            //キーの本来の処理を
            //させたくないときは、trueを返す
            if ((keyData & Keys.KeyCode) == Keys.Left)
            {
                OnDown();
                return true;
            }
            else if ((keyData & Keys.KeyCode) == Keys.Right)
            {
                OnUp();
                return true;
            }
            else if ((keyData & Keys.KeyCode) == Keys.Up)
            {
                OnUp();
                return true;
            }
            else if ((keyData & Keys.KeyCode) == Keys.Down)
            {
                OnDown();
                return true;
            }
            return base.ProcessDialogKey(keyData);
        }

        void OnUp()
        {
            if (checkBoxRotateX.Checked)
                RotateX += 5;
            if (checkBoxRotateY.Checked)
                RotateY += 5;
            if (checkBoxRotateZ.Checked)
                RotateZ += 5;
            if (checkBoxEyeX.Checked)
                EyeX += 0.1f;
            if (checkBoxEyeY.Checked)
                EyeY += 0.1f;
            if (checkBoxEyeZ.Checked)
                EyeZ += 0.1f;
            if (checkBoxTargetX.Checked)
                TargetX += 0.1f;
            if (checkBoxTargetY.Checked)
                TargetY += 0.1f;
            if (checkBoxTargetZ.Checked)
                TargetZ += 0.1f;

            glControl.Refresh();
        }

        void OnDown()
        {
            if (checkBoxRotateX.Checked)
                RotateX -= 5;
            if (checkBoxRotateY.Checked)
                RotateY -= 5;
            if (checkBoxRotateZ.Checked)
                RotateZ -= 5;
            if (checkBoxEyeX.Checked)
                EyeX -= 0.1f;
            if (checkBoxEyeY.Checked)
                EyeY -= 0.1f;
            if (checkBoxEyeZ.Checked)
                EyeZ -= 0.1f;
            if (checkBoxTargetX.Checked)
                TargetX -= 0.1f;
            if (checkBoxTargetY.Checked)
                TargetY -= 0.1f;
            if (checkBoxTargetZ.Checked)
                TargetZ -= 0.1f;

            glControl.Refresh();
        }
    }

    public class Tank
    {
        public Tank()
        {

        }

        // 戦車の位置
        public float X
        {
            get;
            set;
        }
        = 0;

        public float Y
        {
            get;
            set;
        }
        = 0;

        public float Z
        {
            get;
            set;
        }
        = 0;

        // 戦車のX軸を中心とした回転角度
        public float RotateX
        {
            get;
            set;
        }
        = 0;

        // 戦車のY軸を中心とした回転角度
        public float RotateY
        {
            get;
            set;
        }
        = 0;

        // 戦車のZ軸を中心とした回転角度
        public float RotateZ
        {
            get;
            set;
        }
        = 0;

        public Color Color
        {
            get;
            set;
        }

        public void Draw()
        {
            GL.PushMatrix();
            {
                GL.Rotate(RotateX, 1, 0, 0);
                GL.Rotate(RotateY, 0, 1, 0);
                GL.Rotate(RotateZ, 0, 0, 1);
                Draw0();
            }
            GL.PopMatrix();
        }

        public void Draw0()
        {
            GL.Material(MaterialFace.Front, MaterialParameter.Ambient, Color);
            GL.Material(MaterialFace.Front, MaterialParameter.Diffuse, Color);

            GL.Begin(BeginMode.Quads);
            {
                // 左
                GL.Normal3(-Vector3.UnitX);
                GL.Vertex3(-0.25, 0.5, 0.35);
                GL.Vertex3(-0.25, 0.25, 0.35);
                GL.Vertex3(-0.25, 0.25, -0.35);
                GL.Vertex3(-0.25, 0.5, -0.35);

                // 右
                GL.Normal3(Vector3.UnitX);
                GL.Vertex3(0.25, 0.5, 0.35);
                GL.Vertex3(0.25, 0.25, 0.35);
                GL.Vertex3(0.25, 0.25, -0.35);
                GL.Vertex3(0.25, 0.5, -0.35);

                // 正面
                GL.Normal3(Vector3.UnitZ);
                GL.Vertex3(0.25, 0.5, 0.35);
                GL.Vertex3(0.25, 0.25, 0.35);
                GL.Vertex3(-0.25, 0.25, 0.35);
                GL.Vertex3(-0.25, 0.5, 0.35);

                // 背面
                GL.Normal3(-Vector3.UnitZ);
                GL.Vertex3(0.25, 0.5, -0.35);
                GL.Vertex3(0.25, 0.25, -0.35);
                GL.Vertex3(-0.25, 0.25, -0.35);
                GL.Vertex3(-0.25, 0.5, -0.35);

                // 上面
                GL.Normal3(Vector3.UnitY);
                GL.Vertex3(0.25, 0.5, 0.35);
                GL.Vertex3(-0.25, 0.5, 0.35);
                GL.Vertex3(-0.25, 0.5, -0.35);
                GL.Vertex3(0.25, 0.5, -0.35);

                // 砲塔左面
                GL.Normal3(-Vector3.UnitX);
                GL.Vertex3(-0.15, 0.7, 0.1);
                GL.Vertex3(-0.15, 0.5, 0.1);
                GL.Vertex3(-0.15, 0.5, -0.2);
                GL.Vertex3(-0.15, 0.7, -0.2);

                // 砲塔右面
                GL.Normal3(Vector3.UnitX);
                GL.Vertex3(0.15, 0.7, 0.1);
                GL.Vertex3(0.15, 0.5, 0.1);
                GL.Vertex3(0.15, 0.5, -0.2);
                GL.Vertex3(0.15, 0.7, -0.2);

                // 砲塔正面
                GL.Normal3(Vector3.UnitZ);
                GL.Vertex3(-0.15, 0.7, 0.1);
                GL.Vertex3(0.15, 0.7, 0.1);
                GL.Vertex3(0.15, 0.5, 0.1);
                GL.Vertex3(-0.15, 0.5, 0.1);

                // 砲塔背面
                GL.Normal3(-Vector3.UnitZ);
                GL.Vertex3(-0.15, 0.7, -0.2);
                GL.Vertex3(0.15, 0.7, -0.2);
                GL.Vertex3(0.15, 0.5, -0.2);
                GL.Vertex3(-0.15, 0.5, -0.2);

                // 砲塔上面
                GL.Normal3(Vector3.UnitY);
                GL.Vertex3(-0.15, 0.7, 0.1);
                GL.Vertex3(0.15, 0.7, 0.1);
                GL.Vertex3(0.15, 0.7, -0.2);
                GL.Vertex3(-0.15, 0.7, -0.2);

                // 車体
                GL.Normal3(Vector3.UnitY);
                GL.Vertex3(0.35, 0.25, 0.5);
                GL.Vertex3(-0.35, 0.25, 0.5);
                GL.Vertex3(-0.35, 0.25, -0.5);
                GL.Vertex3(0.35, 0.25, -0.5);

                // 砲身
                int n = 16;
                for (int i = 0; i < n; i++)
                {
                    Vector3d vec = new Vector3d(Cos(2 * PI / n * i), Sin(2 * PI / n * i), 0);
                    GL.Normal3(vec);
                    GL.Vertex3(Cos(2 * PI / n * i) * 0.07, Sin(2 * PI / n * i) * 0.07 + 0.6, 0.5);
                    GL.Vertex3(Cos(2 * PI / n * (i + 1)) * 0.07, Sin(2 * PI / n * (i + 1)) * 0.07 + 0.6, 0.5);
                    GL.Vertex3(Cos(2 * PI / n * (i + 1)) * 0.07, Sin(2 * PI / n * (i + 1)) * 0.07 + 0.6, 0.1);
                    GL.Vertex3(Cos(2 * PI / n * i) * 0.07, Sin(2 * PI / n * i) * 0.07 + 0.6, 0.1);
                }
            }
            GL.End();

            // キャタピラ
            GL.Begin(BeginMode.Quads);
            {
                for (int i = 0; i < 12; i++)
                {
                    if (i >= 3 && i <= 8)
                        continue;

                    double r = 0.25 / 2;
                    double z = 0.5 - r;
                    Vector3d vec = new Vector3d(Cos(2 * PI / 12 * i), Sin(2 * PI / 12 * i), 0);
                    GL.Normal3(vec);
                    GL.Vertex3(0.35, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
                    GL.Vertex3(0.15, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
                    GL.Vertex3(0.15, Sin(2 * PI / 12 * (i + 1)) * r + r, Cos(2 * PI / 12 * (i + 1)) * r + z);
                    GL.Vertex3(0.35, Sin(2 * PI / 12 * (i + 1)) * r + r, Cos(2 * PI / 12 * (i + 1)) * r + z);
                }
                for (int i = 0; i < 12; i++)
                {
                    if (i >= 3 && i <= 8)
                        continue;

                    double r = 0.25 / 2;
                    double z = 0.5 - r;
                    Vector3d vec = new Vector3d(Cos(2 * PI / 12 * i), Sin(2 * PI / 12 * i), 0);
                    GL.Normal3(vec);
                    GL.Vertex3(-0.35, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
                    GL.Vertex3(-0.15, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
                    GL.Vertex3(-0.15, Sin(2 * PI / 12 * (i + 1)) * r + r, Cos(2 * PI / 12 * (i + 1)) * r + z);
                    GL.Vertex3(-0.35, Sin(2 * PI / 12 * (i + 1)) * r + r, Cos(2 * PI / 12 * (i + 1)) * r + z);
                }
                for (int i = 0; i < 12; i++)
                {
                    if (i >= 3 && i <= 8)
                    {
                        double r = 0.25 / 2;
                        double z = -0.5 + r;
                        Vector3d vec = new Vector3d(Cos(2 * PI / 12 * i), Sin(2 * PI / 12 * i), 0);
                        GL.Normal3(vec);
                        GL.Vertex3(0.35, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
                        GL.Vertex3(0.15, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
                        GL.Vertex3(0.15, Sin(2 * PI / 12 * (i + 1)) * r + r, Cos(2 * PI / 12 * (i + 1)) * r + z);
                        GL.Vertex3(0.35, Sin(2 * PI / 12 * (i + 1)) * r + r, Cos(2 * PI / 12 * (i + 1)) * r + z);
                    }
                }
                for (int i = 0; i < 12; i++)
                {
                    if (i >= 3 && i <= 8)
                    {
                        double r = 0.25 / 2;
                        double z = -0.5 + r;
                        Vector3d vec = new Vector3d(Cos(2 * PI / 12 * i), Sin(2 * PI / 12 * i), 0);
                        GL.Normal3(vec);
                        GL.Vertex3(-0.35, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
                        GL.Vertex3(-0.15, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
                        GL.Vertex3(-0.15, Sin(2 * PI / 12 * (i + 1)) * r + r, Cos(2 * PI / 12 * (i + 1)) * r + z);
                        GL.Vertex3(-0.35, Sin(2 * PI / 12 * (i + 1)) * r + r, Cos(2 * PI / 12 * (i + 1)) * r + z);
                    }
                }
                GL.Normal3(-Vector3.UnitY);
                GL.Vertex3(0.35, 0, 0.35);
                GL.Vertex3(0.15, 0, 0.35);
                GL.Vertex3(0.15, 0, -0.35);
                GL.Vertex3(0.35, 0, -0.35);

                GL.Normal3(-Vector3.UnitY);
                GL.Vertex3(-0.35, 0, 0.35);
                GL.Vertex3(-0.15, 0, 0.35);
                GL.Vertex3(-0.15, 0, -0.35);
                GL.Vertex3(-0.35, 0, -0.35);
            }
            GL.End();

            // キャタピラ内部の車輪
            GL.Begin(BeginMode.TriangleFan);
            {
                GL.Normal3(Vector3.UnitX);

                double r = 0.25 / 2;
                double z = 0.5 - r;

                GL.Vertex3(0.3, r, z);
                for (int i = 0; i <= 12; i++)
                    GL.Vertex3(0.3, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
            }
            GL.End();
            GL.Begin(BeginMode.TriangleFan);
            {
                GL.Normal3(Vector3.UnitX);

                double r = 0.25 / 2;
                double z = 0.5 - r - r * 3;

                GL.Vertex3(0.3, r, z);
                for (int i = 0; i <= 12; i++)
                    GL.Vertex3(0.3, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
            }
            GL.End();
            GL.Begin(BeginMode.TriangleFan);
            {
                GL.Normal3(Vector3.UnitX);

                double r = 0.25 / 2;
                double z = -0.5 + r;

                GL.Vertex3(0.3, r, z);
                for (int i = 0; i <= 12; i++)
                    GL.Vertex3(0.3, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
            }
            GL.End();
            GL.Begin(BeginMode.TriangleFan);
            {
                GL.Normal3(-Vector3.UnitX);

                double r = 0.25 / 2;
                double z = 0.5 - r;

                GL.Vertex3(-0.3, r, z);
                for (int i = 0; i <= 12; i++)
                    GL.Vertex3(-0.3, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
            }
            GL.End();
            GL.Begin(BeginMode.TriangleFan);
            {
                GL.Normal3(-Vector3.UnitX);

                double r = 0.25 / 2;
                double z = 0.5 - r - r * 3;

                GL.Vertex3(-0.3, r, z);
                for (int i = 0; i <= 12; i++)
                    GL.Vertex3(-0.3, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
            }
            GL.End();
            GL.Begin(BeginMode.TriangleFan);
            {
                GL.Normal3(-Vector3.UnitX);

                double r = 0.25 / 2;
                double z = -0.5 + r;

                GL.Vertex3(-0.3, r, z);
                for (int i = 0; i <= 12; i++)
                    GL.Vertex3(-0.3, Sin(2 * PI / 12 * i) * r + r, Cos(2 * PI / 12 * i) * r + z);
            }
            GL.End();
        }

        double PI
        {
            get { return Math.PI; }
        }

        double Sin(double d)
        {
            return Math.Sin(d);
        }

        double Cos(double d)
        {
            return Math.Cos(d);
        }
    }

    public class GLControlEx : OpenTK.GLControl
    {

    }
}
